function [ beta, Res, evbeta, ev_ALL ] = mmqGM( X, ye, s, model, beta0or, aux, MaxIter, geraX )
% [ beta, Res ] = mmqGM( X, y, s, model, beta0 )
%
%  inspirado na rotina do nlinfit do toolbox do matlab
%  Primeira versao: Otaviano (14-09-2001)
%  Ultima modificacao: Zwinglio (14-10-2014)
%
%  The function is evaluated at yfit = feval(model, beta, X);
%
%  Example:
%  model = @(b,x)( b(1)*x + b(2) );
%  x = 1:100;
%  y = model([1 2],x) + randn(size(x));
%  [ beta, R ] = mmqGM( x, y, 1, model, [1; 2]);
%
%
% New feature introduced in 22-07-2009
%    beta0 can be a cell with { beta0 [betaPrevious inc_betaPrevious] }
%  where betaPrevious (OPTIONAL) is previous knowledge values of each beta and
%  inc_betaPrevious is the incertaint in each of this previous knowledge.
%    inc_betaPrevious(j)=NaN  means that there are no previous knoledge about beta(j)
%    inc_betaPrevious(j)=0    means that beta(j) must be fixed at beta0(j) value.
%  When betaPrevious is absent, betaPrevious = beta0.
%  Example of fit with inc_betaPrevious:
%  model = @(b,x)( b(1)*x + b(2) );
%  x = 1:100;
%  y = model([1 2],x) + randn(size(x));
%  % To fix beta(1)=1 we can use either:
%   [ beta, R ] = mmqGM( x, y, 1, model, {[1;1] [0; nan]} );
%  or:
%   [ beta, R ] = mmqGM( x, y, 1, model, {[1;1] [ [1;1] [0; nan] ]} );
%
%  % To consider a previous knolodge beta(1)=1 with uncertainty 1e-2 we can use:
%   [ beta, R ] = mmqGM( x, y, 1, model, {[1;1] [1e-2; nan]} );
%  or:
%   [ beta, R ] = mmqGM( x, y, 1, model, {[1;1] [ [1;1] [1e-2; nan] ]} );
%
%  The full list of input-output parameters is:
%  [ beta, Res, evbeta, ev_ALL ] = mmqGM( X, ye, s, model, beta0, aux, MaxIter, geraX )
%     with  model = model( beta, X, aux );
%
%  The standard input-output parameters are:
%  [ beta, Res ] = mmqGM( X, y, s, model, beta0 )
%     with  model = model( beta, X );

%  Penultima modificacao: Zwinglio (26-7-2009)

ni_aumento = sqrt(10);
ni_diminuicao = 10;
fatorGM = .01;

% ni_aumento = 1.5;
% ni_diminuicao = 1.5;
% fatorGM = .1;

if nargin<5
    error( 'Non linear fits need intial values of the parameters (beta0)' )
end

ye = ye(:);
n = length(ye);
if size(X,1) ~= n, X = X'; end
[ linX, colX ] = size( X );
% X = X(:);

if length(s)==1
    S = s*ones(n,1);
else
    S = s(:);
end

RELATIONS = [];
if iscell( beta0or )
    beta0 = beta0or{1}(:);
    if length( beta0or )>=2 && ~isempty( beta0or{2} )
        CONSTRAINS = beta0or{2};
        if min( size( CONSTRAINS ) )==1 && max( size( CONSTRAINS ) )==length(beta0)
            CONSTRAINS = [ beta0 CONSTRAINS(:) ];
        end
    else
        CONSTRAINS = repmat( NaN, [length(beta0), 2] );
    end
    if length( beta0or )>=3
        RELATIONS = beta0or{3};
        if size(RELATIONS,2)~=length(beta0)+2
            if size(RELATIONS,1)~=length(beta0)+2
                RELATIONS = RELATIONS';
            else
                RELATIONS = [];
                warning( 'RELATIONS were not considered (wrong size)' )
            end
        end
    end
else
    beta0 = beta0or(:);
    CONSTRAINS = repmat( NaN, [length(beta0), 2] );
end

% CONSTRAINS( CONSTRAINS(:,2)==0, 2 ) = 10*sqrt(eps)*min(S);
% p = length(beta0);
indFREE = find( (isnan(CONSTRAINS(:,2))==1) | (CONSTRAINS(:,2)~=0) );
indFIX = find( CONSTRAINS(:,2)==0 );
pFIX = length( indFIX );
CONSTRAINS( CONSTRAINS(:,2)==0, 2 ) = NaN;
pALL = length(beta0);
p = pALL - pFIX;

indCONSTRAINS = find( isfinite( sum(CONSTRAINS,2) ) );
nCONSTRAINS = sum( isfinite( sum(CONSTRAINS,2) ) );

nRELATIONS = size( RELATIONS, 1 );

indFREE = indFREE(:)';
indFIX = indFIX(:)';
indCONSTRAINS = indCONSTRAINS(:)';

y = ye;

if nCONSTRAINS > 0
    yC = CONSTRAINS(indCONSTRAINS,1);
    SC = CONSTRAINS(indCONSTRAINS,2);
%     indC = length(S)+(1:nCONSTRAINS);
    y = [ y; yC ];
    S = [ S; SC ];
    

    JC = zeros( nCONSTRAINS, pALL );
    for qC = 1 : nCONSTRAINS
        JC( qC, indCONSTRAINS(qC) ) = 1 ./ SC(qC);
    end
end

if nRELATIONS > 0
    yR = RELATIONS(:,1);
    SR = RELATIONS(:,2);
    indR = length(S)+(1:nRELATIONS)';
    
    y = [ y; yR ];
    S = [ S; SR ];
    
    JR_SR = RELATIONS(:,3:end);
end


if nargin>=6 && ~isempty( aux )
    haAux = 1;
else
    haAux = 0;
end

if ~exist( 'MaxIter','var' ) || isempty( MaxIter )
    MaxIter = 100;
end

if ~exist( 'geraX', 'var' ) || isempty( geraX )
    haGeraX = 0;
else
    haGeraX = 1;
end

% J = zeros(n,p);
% J = zeros(n+nCONSTRAINS,p);
beta = beta0;
% betanew = beta + 1 + .1*abs(beta);
betanew = beta + 10*sqrt(eps) + .001*abs(beta);
betanew( indFIX ) = beta( indFIX );
maxiter = MaxIter; % ==============
iter = 0;
betatol = 1.0E-2;
rtol = 1.0E-4;
% qui2 = 1;
% qui2old = qui2;
S = S(:);
v = eye(p);
convergiu = 0;
% disp( 'new fit' )
usandoGM = 0;
while and( convergiu<2, iter < maxiter )
    iter = iter + 1;    
    if iter > 1
        beta = betanew;
        yfit = yfitnew;
    else
        if haAux
            yfit = feval(model, beta, X, aux);
        else
            yfit = feval(model, beta, X);
        end
        if nCONSTRAINS>0
            yfitC = [ beta(indCONSTRAINS) ];
            yfit = [ yfit(:); yfitC(:) ];
        end
        if nRELATIONS>0
            yfitR = JR_SR * beta(:);
            yfit = [ yfit(:); yfitR ];
        end        
    end
    if haGeraX
        if haAux
            J = feval( geraX, beta, X, aux );
        else
            J = feval( geraX, beta, X );
        end
        J = J ./ (S*ones(1,size(J,2)));
    else
        J = zeros(n,pALL);
        for k = indFREE(:)'
            delta = zeros( size( beta) );
            delta(k) = 1e-4 * beta(k);
%             if abs(delta(k)) < 100*eps
%                 nb = sqrt(norm(beta));
%                 delta(k) = sqrt(eps) * (nb + (abs(nb)<(100*sqrt(eps))) );
%             end
            if abs(delta(k)) < 10*sqrt(eps)
                delta(k) = 10*sqrt(eps);
            end
            betaplus = beta + delta;
            if haAux
                yplus = feval(model, betaplus, X, aux);
            else
                yplus = feval(model, betaplus, X);                
            end
            J(:,k) = ( ( yplus(1:n) - yfit(1:n) ) ./ S(1:n) ) / (betaplus(k)-beta(k));
        end
    end
    % to remove FIX
    J = J(:,indFREE );
    
    if nRELATIONS>0
        if sum( SR<0 )
            indSRneg = find( SR<0 );
            JJ = J'*J;
            v = JJ\eye(p);
            s_inter = sqrt( diag( JR_SR(:,indFREE) * v * JR_SR(:,indFREE)' ) );
            S( indR(indSRneg) ) = abs( SR(indSRneg) ) .* s_inter(indSRneg);
        end
        for qR = nRELATIONS:-1:1
            JR_now(qR,:) = JR_SR(qR,:) / S(indR(qR));
        end
    end
    
    % to add constrains
    if nCONSTRAINS
        J = [ J; JC(:,indFREE) ];
    end
    % to add relations
    if nRELATIONS
        J = [ J; JR_now(:,indFREE) ];
    end

    r = ( y(:) - yfit(:) ) ./ S;
    qui2old = r'*r;
    
    
    
    JJ = J'*J;
    reduziu = 0;
    aumentouGM = 0;
    while( reduziu==0 )
        JJplus = JJ + (diag(diag(JJ)))*fatorGM;
        step = (JJplus)\(J'*r);        
        betanew(indFREE) = beta(indFREE) + step;
        if haAux
            yfitnew = feval(model, betanew, X, aux);
        else
            yfitnew = feval(model, betanew, X);
        end
        if nCONSTRAINS>0
            yfitnewC = [ betanew(indCONSTRAINS) ];
            yfitnew = [ yfitnew(:); yfitnewC(:) ];
        end
        if nRELATIONS>0
            yfitnewR = JR_SR * betanew(:);
            yfitnew = [ yfitnew(:); yfitnewR ];
        end

        rnew = (y(:) - yfitnew(:))./S;
        qui2 = rnew'*rnew;
        if ( qui2 > qui2old*(1+1E-4))
            if fatorGM<=1E2
                fatorGM = fatorGM * ni_aumento;
                aumentouGM = aumentouGM + 1;                
            else
                reduziu = -1;
            end
            convergiu = 0;
        else
            reduziu = 1;
        end
    end
    fatorGM = min( fatorGM, 1 );
    if aumentouGM==0
        fatorGM = fatorGM / ni_diminuicao;
        if ( ( abs( qui2old - qui2 ) <= 1E-2*(qui2+qui2old) ) || (qui2<=1e2*eps) ) && fatorGM<.1
            v = ( JJ )\eye(p);
            if convergiu == 0
                convergiu = 1;
                fatorGM = 1E-6;
            elseif convergiu == 1
                if max( abs( step )./sqrt( diag(v)*(qui2/(n+nCONSTRAINS+nRELATIONS-p)) ) )<.05
                    convergiu = 2;
                end
            end
        end
    end
    if nargout>=3
        evbeta(:,iter) = [betanew(:); step(:)];
        if nargout>=4
            ev_ALL(:,iter) = [ log10(fatorGM); aumentouGM; reduziu; qui2old; qui2; ...
                abs( qui2old - qui2 ); 1E-2*(qui2+qui2old); ...
                max( abs( step )./sqrt( diag(v)*(qui2/(n+nCONSTRAINS+nRELATIONS-p)) ) ) ];
        end
    end
end
if convergiu<2, v = ( JJ )\eye(p); end
if iter == maxiter
    warning('mmqGM did not converge. Returning values of the last iteration.');
end

% added in 14/10/10 (to consider the last changes in beta)
beta = betanew;
yfit = yfitnew;

y = y(1:n);
yfit = yfit(1:n);
r = r(1:n);
if pFIX>0
    v2 = v;
    v = zeros( pALL, pALL );
    v(indFREE,indFREE) = v2;        
end

ngl = n+nCONSTRAINS+nRELATIONS-p;
if ngl>0
    try
        Pqui = 1 - chi2cdf( qui2 , ngl );
    catch
        Pqui = NaN;
    end
else
    Pqui = NaN;
end

if nargout>=2
    Res.A = beta;
    Res.SA = sqrt( diag(v) );
    Res.sA = Res.SA;
    Res.VA = v;
    Res.rhoA = v./(Res.SA*Res.SA');
    Res.qui2 = qui2;
    Res.Pqui = Pqui;
    Res.ngl = ngl;
    Res.F = yfit;
    Res.Yfit = yfit;
    Res.iter = iter;
    if length(s)==1
        Res.si = repmat(s,size(y));
    else
        Res.si = s;
    end
    Res.s = s;
    Res.x = X;
    Res.y = y;
    Res.Dif = y - yfit;
    Res.Res = r;
end
